# flake8: noqa E402, F401

import warnings
import sys

# sys.stderr = lambda x: None

warnings.warn = lambda *args, **kwargs: None
warnings._showwarnmsg = lambda x: None

# warnings.simplefilter("ignore")

from rich.console import Console

console = Console()
console._force_terminal = True

import envo.e2e

from envo import e2e
from envo.logs import logger
from envo.devops import *
from envo.env import *
from envo.plugins import *
from envo.misc import EnvoError, import_from_file, add_source_roots, get_repo_root
from envo import venv_utils
from envium import (
    env_var,
    ctx_var,
    computed_env_var,
    computed_secret,
    computed_ctx_var,
    EnvGroup,
    CtxGroup,
    SecretsGroup,
    secret,
)



import os
import sys
from pathlib import Path
from typing import List, Optional, Union

from envo import Env, Namespace, logger

__all__ = [
    "Plugin",
    "VirtualEnv",
]

import envium
from envium import Ctx, CtxGroup, Environ, ctx_var, env_var

from envo import venv_utils
from envo.env import BaseEnv
from envo.logs import Logger
from envo.misc import is_windows


class Plugin(BaseEnv):
    class Environ(Environ):
        ...

    class Ctx(Ctx):
        ...


class VirtualEnv(Plugin):
    """
    Env that activates virtual environment.
    """

    class Environ(Plugin.Environ):
        path: List[str] = env_var(raw=True)

    class Ctx(Plugin.Ctx):
        class Venv(CtxGroup):
            dir: Optional[Path] = ctx_var()
            name: str = ctx_var(".venv")
            discover: bool = ctx_var(False)

        venv = Venv()

    ctx: Ctx
    e: Environ

    def init(self):
        super().init()

        self.__logger: Logger = logger.create_child("venv", descriptor="VirtualEnv")
        self.__logger.debug("VirtualEnv plugin init")

    def post_init(self) -> None:
        if not self.ctx.venv.dir:
            self.ctx.venv.dir = self.meta.root

        try:
            if self.ctx.venv.discover:
                self._venv = venv_utils.DiscoveredVenv(root=self.meta.root, venv_name=self.ctx.venv.name)
            else:
                self._venv = venv_utils.Venv(self.ctx.venv.dir / self.ctx.venv.name)
            self._venv.activate(self.e)
        except venv_utils.CantFindVenv:
            self.__logger.debug("Couldn't find venv. Falling back to predicting")
            self._venv = venv_utils.PredictedVenv(root=self.meta.root, venv_name=self.ctx.venv.name)
            self._venv.activate(self.e)



from pathlib import Path
from typing import TYPE_CHECKING, Type

from envo import misc

if TYPE_CHECKING:
    from envo import Env


class StubGen:
    env: "Env"

    def __init__(self, env: "Env"):
        self.env = env

    def generate(self) -> None:
        if not self.env.get_user_envs():
            return

        self._generate_env(self.env.__class__)
        for p in self.env._parents:
            self._generate_env(p)

    def _generate_env(self, env: Type["Env"]):
        env_descr = misc.EnvParser(env.get_env_path())

        file = Path(f"{env.get_env_path()}i")
        file.touch(exist_ok=True)
        file.write_text(env_descr.get_stub(), "utf-8")



from dataclasses import dataclass

from envo import logger
from envo.misc import Callback


class Status:
    @dataclass
    class Callbacks:
        on_ready: Callback
        on_not_ready: Callback

    _context_ready: bool
    _reloader_ready: bool
    _source_ready: bool

    def __init__(self, calls: Callbacks) -> None:
        self.calls = calls
        self._context_ready = False
        self._reloader_ready = False
        self._source_ready = False

    def __repr__(self) -> str:
        return (
            f"Status (context_ready={self.shell_context_ready},"
            f" reloader_ready={self._reloader_ready}, source_ready={self._source_ready})"
        )

    @property
    def shell_context_ready(self) -> bool:
        return self._context_ready

    @shell_context_ready.setter
    def shell_context_ready(self, value: bool) -> None:
        self._context_ready = value
        logger.debug("Context", {"ready": value})
        self._on_status_change()

    @property
    def source_ready(self) -> bool:
        return self._source_ready

    @source_ready.setter
    def source_ready(self, value: bool) -> None:
        self._source_ready = value
        logger.debug("Source", {"ready": value})
        self._on_status_change()

    @property
    def reloader_ready(self) -> bool:
        return self._reloader_ready

    @reloader_ready.setter
    def reloader_ready(self, value: bool) -> None:
        self._reloader_ready = value
        logger.debug("Reloader", {"ready": value})
        self._on_status_change()

    @property
    def ready(self) -> bool:
        return self.shell_context_ready and self.reloader_ready and self.source_ready

    def _on_status_change(self) -> None:
        if self.ready:
            logger.debug("Everything ready")
            self.calls.on_ready()
        else:
            logger.debug(f"Not ready {repr(self)}")
            self.calls.on_not_ready()



import os
import sys
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union

import envium
from envium import env_var

from envo import Env, Namespace, logger
from envo.misc import is_windows


class CantFindVenv(Exception):
    pass


@dataclass
class VenvPath:
    root_path: Path
    venv_name: str

    @property
    def path(self) -> Path:
        return self.root_path / self.venv_name

    @property
    def lib_path(self) -> Path:
        return self.path / "lib"

    @property
    def bin_path(self) -> Path:
        if is_windows():
            return self.path / "Scripts"
        else:
            return self.path / "bin"

    @property
    def site_packages_path(self) -> Path:
        if is_windows():
            ret = self.lib_path / "site-packages"
            return ret
        try:
            ret = next(self.lib_path.glob("*"))
            # on windows there is no python.* in site packages path
            if ret.name != "site-packages":
                ret /= "site-packages"
            return ret
        except StopIteration:
            raise CantFindVenv()

    @property
    def possible_site_packages(self) -> List[Path]:
        if is_windows():
            return [self.site_packages_path]
        versions = [f"python3.{v}" for v in range(0, 20)]
        versions += [f"python2.{v}" for v in range(0, 10)]
        ret = [self.lib_path / v / "site-packages" for v in versions]
        return ret


@dataclass
class BaseVenv:
    venv_path: VenvPath = field(init=False)

    class Environ(envium.Environ):
        path: List[Path] = env_var(raw=True, default_factory=list)

    def __post_init__(self) -> None:
        self.e = self.Environ(name="envo", load=True)

    def _get_path(self) -> str:
        path = f"""{str(self.venv_path.bin_path)}"""
        return path

    def activate(self, e: Optional[Env.Environ] = None) -> None:
        environ = e or self.e

        path = self._get_path()

        if path not in environ.path:
            environ.path.insert(0, path)

        environ.save_to_os_environ()

    def deactivate(self, e: Optional[Env.Environ] = None) -> None:
        environ = e or self.e
        if self._get_path() in environ.path:
            environ.path.remove(self._get_path())

        environ.save_to_os_environ()

    def __enter__(self) -> None:
        self.activate()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.deactivate()


@dataclass
class PredictedVenv(BaseVenv):
    root: Path
    venv_name: str

    def __post_init__(self) -> None:
        super().__post_init__()

        self.venv_path = VenvPath(root_path=self.root, venv_name=self.venv_name)

    def activate(self, e: Optional[Env.Environ] = None) -> None:
        super().activate(e)
        for d in self.venv_path.possible_site_packages:
            if str(d) not in sys.path:
                sys.path.insert(0, str(d))

    def deactivate(self, e: Optional[Env.Environ] = None) -> None:
        super().deactivate(e)
        for d in self.venv_path.possible_site_packages:
            if str(d) in sys.path:
                sys.path.remove(str(d))


@dataclass
class ExistingVenv(BaseVenv):
    def activate(self, e: Optional[Env.Environ] = None) -> None:
        super().activate(e)

        if str(self.venv_path.site_packages_path) not in sys.path:
            sys.path.insert(0, str(self.venv_path.site_packages_path))

    def deactivate(self, e: Optional[Env.Environ] = None) -> None:
        super().deactivate(e)

        try:
            while str(self.venv_path.site_packages_path) in sys.path:
                sys.path.remove(str(self.venv_path.site_packages_path))
        except CantFindVenv:
            pass


@dataclass
class Venv(ExistingVenv):
    path: Union[Path, str]

    def __post_init__(self) -> None:
        super().__post_init__()

        self.path = Path(self.path).absolute()
        self.venv_path = VenvPath(root_path=self.path.parent, venv_name=self.path.stem)


@dataclass
class DiscoveredVenv(ExistingVenv):
    root: Path
    venv_name: str

    def __post_init__(self) -> None:
        super().__post_init__()

        root_path = self._get_venv_dir()

        self.venv_path = VenvPath(root_path=root_path, venv_name=self.venv_name)

        if not self.venv_path.path.exists():
            raise CantFindVenv

    def _get_venv_dir(self) -> Optional[Path]:
        path = self.root

        while path.parent != path:
            if (path / self.venv_name).exists():
                return path
            path = path.parent

        raise CantFindVenv()
